@@ -14,6 +14,13 @@ import colorio               # pip install colorio
 import cv2                   # pip install opencv-python
 
 
+######################################################################################
+#
+#  P U B L I C   A P I
+#
+######################################################################################
+
+
 class Algorithms:
     """
     A collection of ISP algorithms. See help(rawpipe) for documentation.
@@ -270,20 +277,34 @@ class Algorithms:
                 self._vprint(f"{_elapsed(t0)} - applying CCM {sums}: range = {self._minmax(frame)}")
         return frame
 
+    def gamut(self, frame: np.ndarray, mode: str = "ACES", p=1.2) -> np.ndarray:
+        """
+        Compress out-of-gamut (negative) RGB colors into the visible gamut using
+        the ACES gamut mapping algorithm. Per-channel gamut protection thresholds
+        and distance limits are kept at their defaults, but the compression curve
+        slope can be controlled with the exponent p. If mode is not "ACES", the
+        frame is returned untouched.
+        """
+        if mode == "ACES":
+            assert p >= 1, f"compression curve power must be >= 1.0; was {p:.3f}"
+            t0 = time.time()
+            frame = _aces_gamut(frame, power=p)
+            self._vprint(f"{_elapsed(t0)} - gamut mapping [{mode}, p={p:.2f}]: range = {self._minmax(frame)}")
+        return frame
+
     def tonemap(self, frame: np.ndarray, mode: str = "Reinhard") -> np.ndarray:
         """
         Apply Reinhard tonemapping on the given RGB frame, compressing the range
         [0, N] to [0, 1]. Negative values are clipped to zero. This algorithm is
         format-agnostic. If mode is not "Reinhard", the frame is returned untouched.
         """
-        if mode is not None:
-            if mode == "Reinhard":
-                frame = self.clip(frame, 0, np.inf)
-                t0 = time.time()
-                frame = frame.astype(np.float32)  # can't handle float64
-                algo = cv2.createTonemapReinhard(gamma=1.0, intensity=0.0, light_adapt=0.0, color_adapt=0.0)
-                frame = algo.process(frame)
-                self._vprint(f"{_elapsed(t0)} - tonemapping [{mode}]: range = {self._minmax(frame)}")
+        if mode == "Reinhard":
+            frame = self.clip(frame, 0, np.inf)
+            t0 = time.time()
+            frame = frame.astype(np.float32)  # can't handle float64
+            algo = cv2.createTonemapReinhard(gamma=1.0, intensity=0.0, light_adapt=0.0, color_adapt=0.0)
+            frame = algo.process(frame)
+            self._vprint(f"{_elapsed(t0)} - tonemapping [{mode}]: range = {self._minmax(frame)}")
         return frame
 
     def chroma_denoise(self,
@@ -424,6 +445,13 @@ class Algorithms:
         return ""
 
 
+######################################################################################
+#
+#  I N T E R N A L   F U N C T I O N S
+#
+######################################################################################
+
+
 def _transform_srgb_to_hsl(frame: np.ndarray) -> np.ndarray:
     rgb = frame.swapaxes(-1, 0)  # (H, W, 3) ==> (3, W, H)
     rgb = np.clip(rgb, 0.001, 0.999)  # colorio.cs.HSL() can't handle 0.0 and 1.0
@@ -441,6 +469,44 @@ def _transform_hsl_to_srgb(frame: np.ndarray) -> np.ndarray:
     return rgb
 
 
+def _aces_gamut(rgb, threshold=[0.815, 0.803, 0.880], cyan=0.147, magenta=0.264, yellow=0.312, power=1.2):
+
+    def compress(dist, lim, thr, power):
+        # power(p) compression function plot https://www.desmos.com/calculator/54aytu7hek
+        invp = 1 / power
+        s = (lim - thr) / ((((1 - thr) / (lim - thr)) ** -power) - 1) ** invp # calc y=1 intersect
+        with np.errstate(invalid="ignore"):
+            cdist = thr + (dist - thr) / (1 + ((dist - thr) / s) ** power) ** invp  # compress
+            cdist = np.nan_to_num(cdist)  # result for dist <= thr will be NaN
+            cdist[dist < thr] = dist[dist < thr]  # retain original value if below threshold
+            return cdist
+
+    rgb = np.asarray(rgb)
+    threshold = np.asarray(threshold)
+    if not threshold.shape:
+        threshold = np.tile(threshold, 3)
+
+    # thr is the percentage of the core gamut to protect.
+    thr = np.clip(threshold, -np.inf, 0.9999)
+    thr = thr.reshape([1] * (rgb.ndim - 1) + [3])
+
+    # lim is the max distance from the gamut boundary that will be compressed
+    # 0 is a no-op, 1 will compress colors from a distance of 2 from achromatic to the gamut boundary
+    lim = np.array([cyan + 1, magenta + 1, yellow + 1])
+
+    # achromatic axis
+    ach = np.max(rgb, axis=-1, keepdims=True)
+
+    # distance from the achromatic axis for each color component aka inverse rgb ratios
+    dist = np.where(ach == 0.0, 0.0, (ach - rgb) / np.abs(ach))
+
+    # compress distance with user controlled parameterized shaper function
+    cdist = compress(dist, lim, thr, power)
+    crgb = ach - cdist * np.abs(ach)
+    crgb = np.clip(crgb, 0.0, None)
+    return crgb
+
+
 def _elapsed(t0: float) -> str:
     elapsed = (time.time() - t0) * 1000
     elapsed_str = f"{elapsed:8.2f} ms"
